# pylint: skip-file
import asyncio

import unimatrix.ext.secrets as secrets


async def main():
    print (await secrets.fetch('local', 'pki/pkcs/noop.rsa', {}))

    print (await secrets.fetch('config', None, {
        'value': 'Inline secret!'
    }))
    print (await secrets.fetch('azure', 'example-secret', {
        'vault': 'unimatrixdev',
    }))
    print (await secrets.fetch('google', 'EXAMPLE_SECRET', {
        'project': 'unimatrixdev',
    }))


if __name__ == '__main__':
    asyncio.run(main())



"""Declares :class:`LocalProvider`."""
import aiofiles

from .base import Provider


class LocalProvider(Provider):
    """Secret storage implementation that uses the local disk as a
    storage backend.
    """
    schema_class = None

    def get_secret_version_sync(self, secret: 'Secret', version: str) -> bytes:
        with open(secret.name, 'rb') as f:
            return f.read()

    async def get_secret_version(self, secret: 'Secret', version: str =None):
        async with aiofiles.open(secret.name, 'rb') as f:
            return await f.read()



# pylint: skip-file
import asyncio
import os

import ioc
import unimatrix.runtime
import unimatrix.ext.secrets as secrets


if __name__ == '__main__':
    unimatrix.runtime.on.sync('boot')

    secrets.load()

    print(secrets.get_sync('EXAMPLE_SECRET_AZURE'))
    print(secrets.get_sync('EXAMPLE_SECRET_GOOGLE'))



# pylint: skip-file
import asyncio

import unimatrix.ext.secrets as secrets


if __name__ == '__main__':

    coro = secrets.fetch('google', 'EXAMPLE_SECRET', {
        'project': 'unimatrixdev'
    })
    print( asyncio.run(coro) )



"""Declares :class:`Provider`."""
import abc


class Provider(metaclass=abc.ABCMeta):
    """The base class for all provider implementations."""
    __module__ = 'unimatrix.ext.secrets.providers'

    #: The schema class to validate the provider options.
    schema_class = abc.abstractproperty()

    @abc.abstractmethod
    def get_secret_version_sync(self, secret: 'Secret', version: str = None):
        """Return a version for the given :class:`Secret` object."""
        raise NotImplementedError

    @abc.abstractmethod
    async def get_secret_version(self, secret: 'Secret', version: str = None):
        """Return a version for the given :class:`Secret` object."""
        raise NotImplementedError



# pylint: skip-file
import unimatrix.ext.secrets as secrets


if __name__ == '__main__':
    schema = secrets.SecretSchema()

    print(schema.load({
        'provider': 'azure',
        'name': 'example-secret',
        'opts': {
            'vault': 'unimatrixdev',
        }
    }))

    print(schema.load({
        'provider': 'config',
        'name': None,
        'opts': {
            'value': "Inline secret!"
        }
    }))

    print(schema.load({
        'provider': 'google',
        'name': 'EXAMPLE_SECRET',
        'opts': {
            'project': 'unimatrixdev',
        }
    }))



# pylint: skip-file
import asyncio

import unimatrix.runtime
import unimatrix.ext.secrets as secrets


if __name__ == '__main__':
    asyncio.run(unimatrix.runtime.on('boot'))

    secrets.add({
        'provider': 'local',
        'name': 'pki/pkcs/noop.rsa',
        'alias': 'EXAMPLE_SECRET',
        'encoding': 'utf-8'
    })

    print(asyncio.run(secrets.get('EXAMPLE_SECRET')))



"""Declares :class:`SecretSchema`."""
import marshmallow
import marshmallow.fields

from .providers import get_valid_providers
from .providers import validate_opts
from .secret import Secret


class SecretSchema(marshmallow.Schema):
    """Validation schema to load secrets using providers."""
    provider = marshmallow.fields.String(
        required=True,
        validate=marshmallow.validate.OneOf(get_valid_providers())
    )

    name = marshmallow.fields.String(
        missing=None
    )

    alias = marshmallow.fields.String(
        required=False
    )

    encoding = marshmallow.fields.String(
        required=False,
        validate=marshmallow.validate.OneOf(['utf-8'])
    )

    opts = marshmallow.fields.Dict(
        required=False,
        missing=dict,
        default=dict,
        allow_none=True
    )

    def load(self, *args, **kwargs):
        dto = super().load(*args, **kwargs)
        dto['opts'] = validate_opts(dto['provider'], dto.get('opts') or {})
        return Secret(**dto)



"""Declares :class:`AzureProvider`."""
import marshmallow.fields
from azure.identity.aio import DefaultAzureCredential\
    as AsyncDefaultAzureCredential
from azure.keyvault.secrets.aio import SecretClient\
    as AsyncSecretClient
from azure.identity import DefaultAzureCredential
from azure.keyvault.secrets import SecretClient

from .base import Provider
from .schema import BaseSchema


class AzureProvider(Provider):
    """Secret storage implementation for Azure Key Vault."""
    credential_sync: DefaultAzureCredential = DefaultAzureCredential()
    clients_sync: dict = {}
    encoding = 'utf-8'

    class schema_class(BaseSchema):
        vault = marshmallow.fields.String(required=True)

    def get_sync_client(self, secret):
        """Return a synchronous Azure Key Vault client for the given
        vault.
        """
        vault = secret.opts.vault
        url = f"https://{vault}.vault.azure.net"
        if vault not in self.clients_sync:
            self.clients_sync[vault] = SecretClient(
                vault_url=url,
                credential=self.credential_sync
            )
        return self.clients_sync[vault]

    def get_secret_version_sync(self, secret: 'Secret', version: str =None):
        client = self.get_sync_client(secret)
        secret = client.get_secret(secret.name)
        return str.encode(secret.value, self.encoding)

    async def get_secret_version(self, secret: 'Secret', version: str =None):
        vault = secret.opts.vault
        url = f"https://{vault}.vault.azure.net"
        credential = AsyncDefaultAzureCredential()
        client = AsyncSecretClient(vault_url=url, credential=credential)
        async with client:
            async with credential:
                secret = await client.get_secret(secret.name)
        return str.encode(secret.value, self.encoding)



"""Declares :class:`BaseSchema`."""
import marshmallow
from unimatrix.lib.datastructures import ImmutableDTO


class BaseSchema(marshmallow.Schema):

    def load(self, *args, **kwargs):
        dto = super().load(*args, **kwargs)
        return ImmutableDTO.fromdict(dto)\
            if not self.many\
            else [ImmutableDTO.fromdict(x) for x in dto]



"""Declares :class:`GoogleProvider`."""
import marshmallow.fields
from google.cloud import secretmanager

from .base import Provider
from .schema import BaseSchema


class GoogleProvider(Provider):
    """Secret storage implementation for Google Secret Manager."""
    client = secretmanager.SecretManagerServiceAsyncClient()
    client_sync = secretmanager.SecretManagerServiceClient()

    class schema_class(BaseSchema):
        project = marshmallow.fields.String(required=True)

    def get_secret_version_sync(self, secret: 'Secret', version: str) -> bytes:
        response = self.client_sync.access_secret_version(
            name=self.client_sync.secret_version_path(
                secret.opts.project, secret.name, version or 'latest')
        )
        return response.payload.data

    async def get_secret_version(self, secret: 'Secret', version: str =None):
        response = await self.client.access_secret_version(
            name=self.client.secret_version_path(
                secret.opts.project, secret.name, version or 'latest')
        )
        return response.payload.data



# pylint: skip-file
import asyncio
import os

import ioc
import unimatrix.runtime
import unimatrix.ext.secrets as secrets


if __name__ == '__main__':
    unimatrix.runtime.on.sync('boot')

    secrets.add({
        'provider': 'google',
        'name': 'EXAMPLE_SECRET',
        'alias': 'EXAMPLE_SECRET_GOOGLE',
        'encoding': 'utf-8',
        'opts': {
            'project': 'unimatrixdev',
        }
    })
    secrets.add({
        'provider': 'azure',
        'name': 'example-secret',
        'alias': 'EXAMPLE_SECRET_AZURE',
        'opts': {
            'vault': 'unimatrixdev',
        }
    })

    print(secrets.get_sync('EXAMPLE_SECRET_AZURE'))
    print(secrets.get_sync('EXAMPLE_SECRET_GOOGLE'))



"""Provides functions to configure secrets."""
from unimatrix.lib import etc

from .registry import registry


def load():
    """Load secret configurations from editable text."""
    for alias, params in dict.items(etc.load('secrets.conf') or {}):
        registry.add({**params, 'alias': alias})



# pylint: skip-file
from .conf import load
from .field import SecretField
from .schema import SecretSchema
from .secret import Secret
from .registry import registry


__all__ = [
    'add',
    'fetch',
    'get',
    'get_sync',
    'load',
    'SecretField',
    'SecretSchema',
]


def add(params):
    """Adds a new secret to the secret registry."""
    registry.add(params)


async def fetch(provider: str, name: str, opts: dict, **kwargs) -> bytes:
    """Fetch secret `name` from the secret storage backend using the given
    provider `provider` with given options `opts`.
    """
    secret = Secret(provider, name, opts, **kwargs)
    return await secret.get(None)


def get_sync(name, version=None):
    """Retrieve a secret from the registry. Optionally
    `version` may be specified to indicate a version. If
    the `version` argument is omitted, then the latest
    version is returned.
    """
    return registry[name][version]


async def get(name, version=None):
    """Retrieve a secret from the registry. Optionally
    `version` may be specified to indicate a version. If
    the `version` argument is omitted, then the latest
    version is returned.
    """
    return await registry.get(name, version)



"""Declares :class:`Registry`."""
import logging

from .secret import Secret


class Registry:
    """Provides an interface to store secret in the local
    memory, and to lookup.
    """
    __module__ = 'unimatrix.ext.secrets'
    logger = logging.getLogger('unimatrix.ext.secrets')

    def __init__(self):
        self.__secrets = {}

    def add(self, params):
        """Add a secret to the registry."""
        secret = Secret(**params)
        self.__secrets[secret.alias or secret.name] = secret
        self.logger.debug(
            "Configured secret (provider: %s, name: %s, alias: %s)",
            secret.provider, secret.name, secret.alias or secret.name
        )

    async def get(self, name: str, version: str):
        """Asynchronously retrieve the secret content from its storage
        backend.
        """
        return await self.__secrets[name].get(version)

    def __getitem__(self, name):
        return self.__secrets[name]


registry = Registry()



# pylint: skip-file
import ioc.loader


__all__ = [
    'get_provider',
    'get_provider_class',
    'get_sync',
    'is_valid_provider',
]


PROVIDER_CLASSES = {
    'azure' : 'unimatrix.ext.secrets.providers.azure.AzureProvider',
    'config': 'unimatrix.ext.secrets.providers.config.ConfigProvider',
    'google': 'unimatrix.ext.secrets.providers.google.GoogleProvider',
    'local' : 'unimatrix.ext.secrets.providers.local.LocalProvider',
}


def is_valid_provider(kind: str) -> bool:
    """Return a boolean indicating if the provider kind is valid."""
    return kind in PROVIDER_CLASSES


def get_valid_providers() -> tuple:
    """Return the list of valid providers."""
    return tuple(PROVIDER_CLASSES.keys())


def get_provider_class(kind: str):
    """Return the provider of the given `kind`."""
    return ioc.loader.import_symbol(PROVIDER_CLASSES[kind])


def get_provider(kind: str):
    """Return a :class:`~unimatrix.ext.crypto.providers.Provider` instance
    of the given `kind`.
    """
    return get_provider_class(kind)()


def get_sync(secret, version=None):
    """Resolve the provider for `secret` :class:`~unimatrix.ext.secrets.Secret`
    and return secret content.
    """
    return get_provider(secret.provider)\
        .get_secret_version_sync(secret, version)


async def get(secret, version=None):
    """Resolve the provider for `secret` :class:`~unimatrix.ext.secrets.Secret`
    and return secret content.
    """
    return await get_provider(secret.provider)\
        .get_secret_version(secret, version)


def validate_opts(kind: str, opts: dict):
    """Use the schema specified on the provider class to validate the options
    configured for a secret.
    """
    Provider = get_provider_class(kind)
    schema = Provider.schema_class()
    return schema.load(opts)



# pylint: skip-file
from unimatrix.lib import etc
from unimatrix.conf import settings

from .conf import load
from .registry import registry


async def on_setup(settings=settings):
    for alias, params in dict.items(getattr(settings, 'SECRETS', {})):
        registry.add({**params, 'alias': alias})

    for alias, params in dict.items(etc.app.get('secrets') or {}):
        registry.add({**params, 'alias': alias})

    load()


async def boot():
    pass



# pylint: skip-file
import asyncio

import unimatrix.ext.secrets as secrets


if __name__ == '__main__':
    secrets.add({
        'provider': 'config',
        'name': 'EXAMPLE_SECRET',
        'opts': {'value': 'Secret from config'}
    })

    print(asyncio.run(secrets.get('EXAMPLE_SECRET')))



"""Declares :class:`ConfigProvider`."""
import marshmallow.fields

from .base import Provider
from .schema import BaseSchema


class ConfigProvider(Provider):
    """Secret storage implementation that uses the in-memory configuration as a
    storage backend.
    """

    class schema_class(BaseSchema):
        value = marshmallow.fields.String(required=True)

    def get_secret_version_sync(self, secret: 'Secret', version: str) -> bytes:
        return str.encode(secret.opts.value, 'utf-8')

    async def get_secret_version(self, secret: 'Secret', version: str =None):
        return str.encode(secret.opts.value, 'utf-8')



# pylint: skip-file
import marshmallow

import unimatrix.ext.secrets as secrets


class ExampleSchema(marshmallow.Schema):
    value = secrets.SecretField()


if __name__ == '__main__':
    schema = ExampleSchema()

    dto = schema.load({
        'value': {
            'provider': 'config',
            'opts': {'value': 'Inline secret'}
        }
    })

    print(dto)



"""Declares :class:`SecretField`."""
import marshmallow.fields

from .schema import SecretSchema


class SecretField(marshmallow.fields.Dict):
    """Field that deserializes the input parameters and uses them to fetch
    a secret from the secret storage backend.
    """

    def _serialize(self, value, attr, obj, **kwargs):
        raise NotImplementedError

    def _deserialize(self, *args, **kwargs):
        schema = SecretSchema()
        secret = schema.load(super()._deserialize(*args, **kwargs))
        return secret[None]
